# Copyright (c) 2002-2004 Infrae. All rights reserved.
# See also LICENSE.txt
# $Revision: 1.28 $

"""Silva extension specifically for the EuroMar conference
"""

# Zope
from OFS import SimpleItem
from AccessControl import ClassSecurityInfo
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
import Globals, zLOG

# Silva
from Products.Silva import SilvaPermissions
from Products.Silva.helpers import add_and_edit
from Products.Silva.interfaces import IContainer
from Products.Silva import mangle
from Products.Formulator.Form import ZMIForm
from Products.Formulator.XMLToForm import XMLToForm
from Products.Formulator.Errors import ValidationError, FormValidationError

# Python
from DateTime import DateTime
import md5, time, os, random, string, urllib
import _mysql_exceptions

class InternalValidationError(Exception):
    """raised when the additional validations fail"""

class FormValidator:
    """form validation code, specific to EUROMAR"""

    show_fields = ['last_name', 'first_name', 'email', 'salutation',
                    'mytitle', 'company', 'department', 'street',
                    'zip', 'city', 'country', 'phone', 'fax', 'comments',
                    'participation', 'conference_package', 'room',
                    'smoking']
    mandatory = ['salutation', 'mytitle', 'last_name', 'first_name',
                    'email', 'street', 'zip', 'city', 'country', 'phone',
                    'participation', 'room']

    def __init__(self, form, stud_date=DateTime(), euro_date=DateTime(), 
                        mari_date=DateTime(), password=None):
        self._form = form
        self._stud_date = stud_date
        self._euro_date = euro_date
        self._mari_date = mari_date
        self._password = password

    def validate(self, request):
        """validate the request

            returns a tuple (messages, show_fields, mandatory) where message
            is a string with all validation messages (errors and hints),
            show_fields a list with the ids of all the fields to render and
            mandatory a list of ids of all fields that are mandatory

            can raise an InternalValidationError when Formulator doesn't like
            the filled-in values
        """
        # let Formulator handle the real validation and raise an exception
        # on truly problematic issues
        try:
            result = self._form.validate_all(request)
        except FormValidationError, e:
            raise InternalValidationError, (self._render_form_errors(e),
                                            self.show_fields[:],
                                            self.mandatory[:])

        show_fields = self.show_fields[:]
        mandatory = self.mandatory[:]
        messages = []

        msg, mnd = self._validate_helper(result)
        if msg:
            messages.append(self._render_additional_errors(msg))
        for id in mnd:
            if id not in show_fields:
                show_fields.append(id)
        mandatory += mnd

        if result['room'][0] == 'd' and not result['mate_last_name']:
            messages.append('Additional information (Preferential room '
                            'mate: first and last name) required for '
                            'double bedroom registrations')
            show_fields += ['mate_last_name', 'mate_first_name']

        return (''.join(messages), show_fields, mandatory)

    def validateRaise(self, request):
        """validates, raises an InternalValidationError when validation
            failed, exception information is a tuple (message, show_fields,
            mandatory), if all goes well it returns the result

            use this to actually do validation before submit, if all goes
            well it doesn't raise an exception, if not it also displays
            what needs to be displayed (rather then just telling you what
            needs to be displayed: this doesn't contain explicit information
            about whether the validation was successful)
        """
        messages = []
        show_fields = self.show_fields[:]
        mandatory = self.mandatory[:]
        try:
            result = self._form.validate_all(request)
        except FormValidationError, e:
            raise InternalValidationError, (self._render_form_errors(e), 
                                                show_fields, mandatory)

        msg, mnd = self._validate_helper(result)
        if msg:
            messages.append(self._render_additional_errors(msg))
        for id in mnd:
            if id not in show_fields:
                show_fields.append(id)
        mandatory += mnd

        if not messages:
            # no failing fields, so validation is correct
            return result

        if result['room'][0] == 'd':
            messages.append('Additional information (Preferential room '
                            'mate: last and first name) desired for '
                            'double bedroom registrations')
            show_fields += ['mate_last_name', 'mate_first_name']

        raise InternalValidationError, (''.join(messages), 
                                            show_fields, mandatory)

    def _validate_helper(self, result):
        dob = result.get('dob', None)
        esr = result.get('esr', None)
        part = result['participation']
        room = result['room']
        cp = result['conference_package']

        messages = []
        additional_mandatory = []
        if part[:4] in ['stud', 'euro', 'mari']:
            if dob is None:
                messages.append('Additional information (date of birth) required '
                                'for the chosen participation type')
            else:
                if ((part[:4] == 'stud' and dob < self._stud_date) or
                        (part[:4] == 'euro' and dob < self._euro_date) or
                        (part[:4] == 'mari' and dob < self._mari_date)):
                    messages.append('Date of birth and participation type '
                                    'not compatible')
            additional_mandatory.append('dob')

        if part[:4] == 'mari':
            if esr != 'yes':
                messages.append('You should confirm to be an Early Stage Researcher for '
                                'chosen participation type')
            additional_mandatory.append('esr')

        if (part[:4] == 'acco' and room[0] == 's'):
                messages.append('Accompanying persons cannot opt for a single room')


        if cp:
            result['mycode']=part[0:4]+'-'+room[0]+'-y'
            if room[0] == 'n':
                messages.append('With a conference package you can only be '
                                'lodged in the conference center')
            if (part.startswith('mari') and room[0] == 's'):
                messages.append('Marie Curie grantees cannot opt for single rooms')
        else:
            result['mycode']=part[0:4]+'-'+room[0]+'-n'
            if part[:4] in ['euro', 'mari']:
                messages.append('Grant applicants must select a conference package')
            else:
                if not result['arrival'] or not result['departure']:
                    messages.append('Additional information required if you do not '
                                    'choose for a conference package')
                else:
                    start = DateTime('2005/07/03')
                    end = DateTime('2005/07/05')
                    arr = DateTime(result['arrival'])
                    dep = DateTime(result['departure'])
                    if arr < start or arr > end:
                        messages.append('Invalid arrival date')
                    if dep < start or dep > end:
                        messages.append('Invalid departure date')
                    if arr > dep:
                        messages.append('Cannot depart before arrival')
                additional_mandatory += ['arrival', 'departure']

        if part[:4] in ['invi', 'conf']:
            if not result['password']:
                messages.append('Password required for speakers and '
                                'conference committee participation '
                                'registration')
            elif result['password'] != self._password:
                messages.append('Invalid participation password')
            additional_mandatory.append('password')

        return (messages, additional_mandatory)

    def _render_form_errors(self, errors):
        ret = ['<ul>']
        for error in errors.errors:
            ret.append('<li>%s: %s</li>' % (error.field['title'], error.error_text))
        ret.append('</ul>')
        return '\n'.join(ret)

    def _render_additional_errors(self, errors):
        ret = ['<ul']
        for error in errors:
            ret.append('<li>%s</li>' % error)
        ret.append('</ul>')
        return '\n'.join(ret)

def sql_quote(data):
    for key, value in data.items():
        if type(value) == bool:
            value = int(value)
        if type(value) not in [str, unicode]:
            value = str(value)
        data[key] = value.replace("'", "''")
    # return the result just to be sure (works in-place!!)
    return data

class ConferenceService(SimpleItem.SimpleItem):
    """A registry to manage conferences

        Uses a MySQL db to get and set data to.

        Specific extension for a conference held by one of our customers.
    """

    security = ClassSecurityInfo()
    meta_type = 'Silva Conference Service'

    _max_validation_period = 7 # in days
    
    _email_address = 'secr@euromar.org'
    
    _validation_email = (
                'Dear Conference Attendee, \n\n'
                'A request has been placed to register you to the '
                'EuroMar conference.\nTo make sure the registration '
                'is valid, you will need to follow\nthe following URL '
                '(some browsers will make it clickable):\n\n'
                '%(baseurl)s/finalizeRegistration?username=%(username)s&'
                'validationkey=%(validationkey)s\n\n'
                'Note that the registration will expire in '
                '%(max_validation)s days.\n\n'
                'Thank you for your interest, \n\n'
                'the conference administration'
                )

    _validation_email_subject = 'EuroMar registration'

    _confirmation_email = (
                'Dear %(first_name)s %(last_name)s,\n\n'
                'Your registration data has successfully been entered for the conference.\n\n'
                'Name: %(salutation)s %(mytitle)s %(first_name)s %(last_name)s\n\n'
                'Department, Company: %(department)s, %(company)s\n'
                '            Address: %(street)s, %(zip)s, %(city)s, %(country)s\n'
                '          Telephone: %(phone)s\n'
                '                Fax: %(fax)s\n'
                '      Email address: %(email)s\n\n'
		'Type of participant: %(participation)s\n'
		'       Room request: %(room)s\n'
		'                     double room preference %(mate_first_name)s %(mate_last_name)s\n'
		' Conference Package: %(conference_package)s\n'
		'            Smoking: %(smoking)s\n'
                '            Due fee: %(fee)s\n\n'
                ' Member publication: %(member_publication)s\n'
                '           Comments: %(comments)s\n\n'
                'An area where you can edit your user data and abstracts has been created\n'
                'on the following URL:\n\n'
                '%(member_publication)s\n\n'
                'To access it (and certain non-public areas of the site), use the following login information:\n\n'
		'User name: your e-mail address (the one to which this email has been sent to.\n'
		'Password: %(password)s \n\n'
		'Note: This e-mail message can never be interpreted as a formal acceptance to the conference.\n'
		'On the basis of the registration data you may receive an invoice.\n'
		'When the due conference fee has been received, you will receive a formal registration letter.\n\n\n'
                'Thanks you for your interest,\n\n'
                'the conference administration'
                )

    _confirmation_email_subject = 'EuroMar conference registration confirmation'

    _admin_email = (
                'A new user has registered:\n\n'
                'Name: %(salutation)s, %(mytitle)s, %(first_name)s, %(last_name)s\n\n'
                'Department, Company: %(department)s, %(company)s\n'
                '            Address: %(street)s, %(zip)s, %(city)s, %(country)s\n'
                '          Telephone: %(phone)s\n'
                '                Fax: %(fax)s\n'
                '      Email address: %(email)s\n\n'
		'Type of participant: %(participation)s\n'
		'       Room request: %(room)s\n'
		'                     double room preference%(mate_first_name)s %(mate_last_name)s\n'
		' Conference Package: %(conference_package)s\n'
		'            Smoking: %(smoking)s\n'
                '            Due fee: %(fee)s\n\n'
                ' Member publication: %(member_publication)s\n\n'
                '           Comments: %(comments)s\n'
                )

    _admin_email_subject = 'EuroMar registration (e-mail backup)'

    # some Euromar specific data that should be editable through the ZMI
    # XXX the default values aren't very useful

    _stud_date = DateTime('1965-07-03')
    _euro_date = DateTime('1975-07-03')
    _mari_date = DateTime('1965-07-03')
    _last_pre_date = DateTime('2005-05-20')

    _invi_comm_password = 'IAmWhoISayIAm'

    _fees = (   ('full-s-y', ('Conference Package: Regular particiant,   Single room', 980)),
                ('full-d-y', ('Conference Package: Regular particiant,   Double room', 880)),
                ('vend-s-y', ('Conference Package: Vendor delegate,      Single room', 750)),
                ('acco-d-y', ('Conference Package: Accompanying person,  Double room', 650)),
                ('stud-s-y', ('Conference Package: Student particiant,   Single room', 805)),
                ('stud-d-y', ('Conference Package: Student particiant,   Double room', 705)),
                ('euro-s-y', ('Conference Package: EuroMAR grant,        Single room', 490)),
                ('euro-d-y', ('Conference Package: EuroMAR grant,        Double room', 390)),
                ('mari-d-y', ('Conference Package: Marie Curie grant,    Double room', 0)),
                ('invi-s-y', ('Conference Package: Invited Speaker,      Single room', 0)),
                ('invi-d-y', ('Conference Package: Invited Speaker,      Double room', 0)),
                ('conf-s-y', ('Conference Package: Conference Committee, Single room', 0)),
                ('conf-d-y', ('Conference Package: Conference Committee, Double room', 0)),
                ('full-x-n', ('    Conference fee: regular participant', 350)),
                ('stud-x-n', ('    Conference fee: student participant', 175)),
                ('acco-x-n', ('    Conference fee: accompanying person', 120)),
                ('board-s',  ('Fee: Full board and lodging,              Single room', 145)),
                ('board-d',  ('Fee: Full board and lodging,              Double room', 125)),
                ('voucher',  ('Fee: Voucher for lunch and dinner (per day)', 50)),
                ('external', ('Registration surcharge when not accomodated in the venue', 25)),
                ('late',     ('Registration surcharge for late registration/payment', 100)),
                )

    
    manage_options = (
        {'label':'Edit', 'action':'manage_editForm'},
        {'label':'View', 'action':'manage_viewForm'},
        ) + SimpleItem.SimpleItem.manage_options

    security.declareProtected(SilvaPermissions.ViewManagementScreens,
                                'manage_editForm')
    manage_editForm = PageTemplateFile(
        'www/serviceConferenceEdit', globals(),  __name__='manage_editForm')

    manage_main = manage_editForm

    edit = manage_editForm

    security.declarePublic('manage_viewForm')
    manage_viewForm = PageTemplateFile(
        'www/serviceConferenceView', globals(),  __name__='manage_viewForm')

    index_html = register = manage_viewForm

    security.declarePublic('manage_viewThanks')
    manage_viewThanks = PageTemplateFile(
        'www/serviceConferenceThanks', globals(), __name__='manage_viewThanks')

    security.declarePublic('manage_viewRegistered')
    manage_viewRegistered = PageTemplateFile(
        'www/serviceConferenceRegistered', globals(), __name__='manage_viewRegistered')

    # provide a Silva-like override mechanism: if there's a macro called 
    # 'override.html' in the 'www' directory, use this as the macro for the 
    # views on the service, else fall back to the nearest 'layout_macro.html'
    # to get the macro see the 'get_view_macro' method below
    security.declarePublic('manage_overrideTemplate')
    manage_overrideTemplate = PageTemplateFile(
        'www/override.html', globals(), __name__='manage_overrideTemplate')

    def __init__(self, id, title):
        self.id = id
        self.title = title
        # mapping with validationkey as key, value is (username, time)
        self._validating = {}
        self.form = None

    def manage_afterAdd(self, *args, **kwargs):
        self._home_publication_location = '/'.join(
                                self.get_root().getPhysicalPath())
        self.load_form()

    security.declarePublic('get_view_macros')
    def get_view_macros(self):
        try:
            macros = self.manage_overrideTemplate.macros
        except IOError:
            macros = getattr(self, 'layout_macro.html').macros
        return macros

    security.declareProtected(SilvaPermissions.ViewManagementScreens,
                                'load_form')
    def load_form(self):
        """load the form

            is a TTW method to allow loading form after updating product code
        """
        # load the form
        self.form = ZMIForm('form', 'RegistrationForm')
        filename = os.path.join(
            Globals.package_home(globals()), 'www', 'registration_form.form')
        f = open(filename)
        XMLToForm(f.read(), self.form)

    security.declareProtected(SilvaPermissions.View,
                                'validateRegistration')
    def validateRegistration(self, REQUEST):
        """Sends an email to an attendee that has attempted to register

            The attendee should respond by clicking a URL that points to
            self.registerAttendee() within self._max_validation_period
            days
        """
        validator = FormValidator(self.form, self._stud_date, 
                                    self._euro_date, self._mari_date,
                                    self._invi_comm_password)
        try:
            message, show_fields, mandatory = validator.validate(REQUEST)
        except InternalValidationError, e:
            return self.manage_viewForm(message=e)
        if not message:
            message = ('Validation successful. Please review the values '
                        'and press \'submit\'.')
        return self.manage_viewForm(message=message, show_fields=show_fields,
                                    mandatory=mandatory)

    security.declareProtected(SilvaPermissions.View,
                                'startRegistration')
    def startRegistration(self, REQUEST):
        """Sends an email to an attendee that has attempted to register

            The attendee should respond by clicking a URL that points to
            self.registerAttendee() within self._max_validation_period
            days
        """
        errors = ''

        validator = FormValidator(self.form, self._stud_date, 
                                    self._euro_date, self._mari_date,
                                    self._invi_comm_password)
        try:
            result = validator.validateRaise(REQUEST)
        except InternalValidationError, e:
            message = e[0]
            show_fields = e[1]
            mandatory = e[2]
            return self.manage_viewForm(message=message, 
                                        show_fields=show_fields,
                                        mandatory=mandatory)

        result['fee'] = result['duefee'] = self._calculate_fee(result)
        result['receivedfee'] = 0

        # XXX dunno password type, domains and realm, we probably want to 
        # retrieve those from mysqlUserFolder (or do we want to use that to 
        # actually create the field and update after that?)
        
        # let's use the user's email address as his/her username for now
        username = result['email']
        
        # large nasty SQL statement, perhaps we want to add the field ids
        # at some point too, I really don't feel like typing those right
        # now though...
        query = ('INSERT INTO %(table)s ('
                    'Username, '
                    'Realname, '
                    'Password, '
                    'PasswordType, '
                    'Domains, '
                    'Realm, '
                    'EMail, '
                    'Salutation, '
                    'Mytitle, '
                    'Lastname, '
                    'Firstname, '
                    'MateLastName, '
                    'MateFirstName, '
                    'Company, '
                    'Department, '
                    'Street, '
                    'Zip, '
                    'City, '
                    'Country, '
                    'Phone, '
                    'Fax, '
                    'Comments, '
                    'DoB, '
                    'Esr, '
                    'DueFee, '
                    'ReceivedFee, '
                    'Arrival, '
                    'Departure, '
                    'Participation, '
                    'ConferencePackage, '
                    'Room, '
                    'Smoking, '
                    'FinalizedRegistration '
                    ') VALUES ('
                    "'%(Username)s', "
                    "'%(Realname)s', "
                    "'%(Password)s', "
                    "'0', "
                    "'', "
                    "'*', "
                    "'%(EMail)s', "
                    "'%(Salutation)s', "
                    "'%(Mytitle)s', "
                    "'%(Lastname)s', "
                    "'%(Firstname)s', "
                    "'%(MateLastName)s', "
                    "'%(MateFirstName)s', "
                    "'%(Company)s', "
                    "'%(Department)s', "
                    "'%(Street)s', "
                    "'%(Zip)s', "
                    "'%(City)s', "
                    "'%(Country)s', "
                    "'%(Phone)s', "
                    "'%(Fax)s', "
                    "'%(Comments)s', "
                    "'%(DoB)s', "
                    "'%(Esr)s', "
                    "'%(DueFee)s', "
                    "'%(ReceivedFee)s', "
                    "'%(Arrival)s', "
                    "'%(Departure)s', "
                    "'%(Participation)s', "
                    "'%(ConferencePackage)s', "
                    "'%(Room)s', "
                    "'%(Smoking)s', "
                    "'0'"
                    ")")

        # escape quotes
        sql_quote(result)

        password = result['password'] = self._create_password()

        data = {'table': 'Users',
                'Username': result['email'],
                'Realname': '%s, %s' % (result['last_name'], result['first_name']),
                'Password': password,
                'EMail': result['email'],
                'Salutation': result['salutation'],
                'Mytitle': result['mytitle'],
                'Lastname': result['last_name'],
                'Firstname': result['first_name'],
                'MateLastName': result['mate_last_name'],
                'MateFirstName': result['mate_first_name'],
                'Company': result['company'],
                'Department': result['department'],
                'Street': result['street'],
                'Zip': result['zip'],
                'City': result['city'],
                'Country': result['country'],
                'Phone': result['phone'],
                'Fax': result['fax'],
                'Comments': result['comments'],
                'DoB': result['dob'],
                'Esr': result['esr'],
                'DueFee': result['duefee'],
                'ReceivedFee': result['receivedfee'],
                'Arrival': result['arrival'],
                'Departure': result['departure'],
                'Participation': result['participation'],
                'ConferencePackage': result['conference_package'],
                'Room': result['room'],
                'Smoking': result['smoking'],
                }

        query = query % data

        # perform the query, the user's data is added to the db
        try:
            self.perform_query(query)
        except _mysql_exceptions.IntegrityError, e:
            id, message = e[0], unicode(e[1], 'ISO-8859-1')
            if id == 1062:
                message = ('Username already in use, it looks like you\'ve '
                            'already used that email address to register')
            return self.manage_viewForm(message=message)

        # create validationkey and send validation email
        validationkey = self.getValidationKey(username)

        # to get the user's entered info without having to load it from the db
        # later on, we store it in the _validating dict too
        self._validating[validationkey] = (username, DateTime(), result)
        email = self._validation_email % {
                    'baseurl': self.absolute_url(),
                    'username': username,
                    'validationkey': validationkey,
                    'max_validation': self._max_validation_period,
                    }
        self._send_email(result['email'], email, self._validation_email_subject)

        # XXX send email to the registrar too?

        return self.manage_viewThanks()

    security.declarePrivate(SilvaPermissions.View,
                                'getValidationKey')
    def getValidationKey(self, username):
        """Generates a unique key that can be used for attendee validation

            the key should be mailed to the user along with a URL that points
            to the 'registerAttendee' method, that URL should have 'username'
            and 'validationkey' as its query variables, where the username
            var should have the id of the attendee (as registered in the MySQL
            database) as value and as the value of validationkey the value 
            returned by this method
        """
        while 1:
            hash = md5.new()
            hash.update(str(time.time()) + username)
            validationkey = hash.hexdigest()
            if validationkey not in self._validating.keys():
                return validationkey

    security.declareProtected(SilvaPermissions.View,
                                'finalizeRegistration')
    def finalizeRegistration(self, username, validationkey):
        """Called on email validation

            creates the attendees Member Publication (to add abstracts etc.)
            and sets it up
        """
        # first remove all entries that are too old
        self._cleanup_validating()
        # see if the user is (still) mentioned in the validating list
        if validationkey not in self._validating.keys():
            return self.manage_viewRegistered(
                    message=('Your name doesn\'t appear on the list, '
                                'most likely because you have confirmed '
                                'your registration already (did you '
                                'accidentally double-click on the link?)'))
        # check integrity a bit
        assert username == self._validating[validationkey][0], \
                    'incorrect validationkey, please re-register'

        # get the data the user entered
        userdata = self._validating[validationkey][2]

        # and remove the validating entry
        del self._validating[validationkey]
        
        # user seems to be genuine, continue setting up his space
        member_pub_url = self._setup_user(username)

        # send registration confirmation to user
        interpolations = {'member_publication': member_pub_url
                          }
        interpolations.update(userdata)

        email = self._confirmation_email % interpolations
        
        self._send_email(userdata['email'], email, 
                          self._confirmation_email_subject)

        email = self._admin_email % interpolations
        
        self._send_email(self._email_address, email, 
                          self._admin_email_subject)

        query = ("UPDATE Users SET FinalizedRegistration = '1' WHERE "
                    "Username = '%s'" % username)

        self.perform_query(query)
        
        return self.manage_viewRegistered()

    security.declareProtected(SilvaPermissions.ViewManagementScreens,
                                'get_email_address')
    def get_email_address(self):
        return self._email_address

    security.declareProtected(SilvaPermissions.ViewManagementScreens,
                                'get_validation_email')
    def get_validation_email(self):
        return self._validation_email

    security.declareProtected(SilvaPermissions.ViewManagementScreens,
                                'get_validation_email_subject')
    def get_validation_email_subject(self):
        return self._validation_email_subject

    security.declareProtected(SilvaPermissions.ViewManagementScreens,
                                'get_confirmation_email')
    def get_confirmation_email(self):
        return self._confirmation_email

    security.declareProtected(SilvaPermissions.ViewManagementScreens,
                                'get_confirmation_email_subject')
    def get_confirmation_email_subject(self):
        return self._confirmation_email_subject

    security.declareProtected(SilvaPermissions.ViewManagementScreens,
                                'get_admin_email')
    def get_admin_email(self):
        return self._admin_email

    security.declareProtected(SilvaPermissions.ViewManagementScreens,
                                'get_admin_email_subject')
    def get_admin_email_subject(self):
        return self._admin_email_subject

    security.declareProtected(SilvaPermissions.ViewManagementScreens,
                                'get_home_publication_location')
    def get_home_publication_location(self):
        return self._home_publication_location

    security.declareProtected(SilvaPermissions.View,
                                'get_max_validation_period')
    def get_max_validation_period(self):
        return self._max_validation_period

    security.declareProtected(SilvaPermissions.ViewManagementScreens,
                                'get_stud_date')
    def get_stud_date(self):
        return self._stud_date

    security.declareProtected(SilvaPermissions.ViewManagementScreens,
                                'get_euro_date')
    def get_euro_date(self):
        return self._euro_date

    security.declareProtected(SilvaPermissions.ViewManagementScreens,
                                'get_mari_date')
    def get_mari_date(self):
        return self._mari_date

    security.declareProtected(SilvaPermissions.ViewManagementScreens,
                                'get_last_pre_date')
    def get_last_pre_date(self):
        return self._last_pre_date

    # the fees list is not editable but let's at least expose it so
    # it can be shown in views...
    security.declareProtected(SilvaPermissions.AccessContentsInformation,
                                'get_fees')
    def get_fees(self):
        return self._fees

    security.declareProtected(SilvaPermissions.ViewManagementScreens,
                                'manage_edit')
    def manage_edit(self, REQUEST):
        """set the object's settings from the ZMI"""
        errors = []
        
        # homepath
        homepath = REQUEST['homepath']
        try:
            location = self.restrictedTraverse(homepath.split('/'))
        except KeyError:
            location = None
        if not location:
            errors.append('path not found')
        elif not IContainer.isImplementedBy(location):
            errors.append('location is not a container type')
        else:
            self._home_publication_location = '/'.join(location.getPhysicalPath())

        # email stuff
        email_address = REQUEST['email_address']
        if not email_address:
            errors.append('no email address entered')
        else:
            self._email_address = email_address

        validation_email = REQUEST['validation_email']
        if not validation_email:
            errors.append('no validation email entered')
        else:
            self._validation_email = validation_email

        validation_email_subject = REQUEST['validation_email_subject']
        if not validation_email_subject:
            errors.append('no validation email subject entered')
        else:
            self._validation_email_subject = validation_email_subject

        confirmation_email = REQUEST['confirmation_email']
        if not confirmation_email:
            errors.append('no confirmation email entered')
        else:
            self._confirmation_email = confirmation_email

        confirmation_email_subject = REQUEST['confirmation_email_subject']
        if not confirmation_email_subject:
            errors.append('no confirmation email subject entered')
        else:
            self._confirmation_email_subject = confirmation_email_subject
        
        admin_email = REQUEST['admin_email']
        if not admin_email:
            errors.append('no admin email entered')
        else:
            self._admin_email = admin_email

        admin_email_subject = REQUEST['admin_email_subject']
        if not admin_email_subject:
            errors.append('no admin email subject entered')
        else:
            self._admin_email_subject = admin_email_subject
        
        # max validation period
        max_validation_period = REQUEST['max_validation_period']
        if not max_validation_period:
            errors.append('no maximum validation period entered')
        else:
            self._max_validation_period = max_validation_period

        try:
            self._stud_date = DateTime(REQUEST['stud_date'])
        except SyntaxError:
            errors.append('student minimum birth date not valid')

        try:
            self._euro_date = DateTime(REQUEST['euro_date'])
        except SyntaxError:
            errors.append('EUROMAR grants minimum birth date not valid')

        try:
            self._mari_date = DateTime(REQUEST['mari_date'])
        except SyntaxError:
            errors.append('Marie Curie grants minimum birth date not valid')

        try:
            self._last_pre_date = DateTime(REQUEST['last_pre_date'])
        except SyntaxError:
            errors.append('Last date of pre-registration not valid')

        if errors:
            return self.manage_editForm(manage_tabs_message='\n'.join(errors))
        else:
            return self.manage_editForm(manage_tabs_message='Data updated successfully')

    def _setup_user(self, username):
        """Create a user publication, RegistrationInformation object, etc."""
        # create Home Publication
        # first make the id valid and unique
        homeid = mangle.Id(self, username)
        homeid.cook()
        homeid.unique()
        homeid = str(homeid)
        
        homes = self.restrictedTraverse(
                        self._home_publication_location.split('/'))
        homes.manage_addProduct['SilvaConferenceExtension'].\
                        manage_addHomePublication(homeid, username)
        home = getattr(homes, homeid)

        # make the user Author in his home folder
        home.manage_addLocalRoles(username, ['Author'])

        self.service_members.get_member(username)
        
        # create RegistrationInformation object
        home.manage_addProduct[
            'SilvaConferenceExtension'
        ].manage_addRegistrationInformation(
            'registrationinfo', 'Registration information', username
        )

        return home.absolute_url()
        
    def _create_homeid_from_username(self, username):
        return urllib.quote(username)

    def _cleanup_validating(self):
        """Removes all validation keys for users that have requested 
            validation too long ago
        """
        now = DateTime()
        for key, (username, time, data) in self._validating.items():
            if now == time > self._max_validation_period:
                # remove the user both from the mapping and the db
                del self._validating[key]
                query = "DELETE FROM Users WHERE Username = '%s'" % username
                self.perform_query(query)

    def perform_query(self, query):
        userfolder = self.acl_users
        return userfolder.rawQuery(query.encode('ISO-8859-1', 'replace'))

    def _create_password(self):
        """create some random password"""
        ret = []
        for i in range(2):
            ret.append(string.digits[random.randrange(len(string.digits))])
        for i in range(3):
            ret.append(string.ascii_lowercase[random.randrange(len(string.ascii_lowercase))])
        for i in range(2):
            ret.append(string.digits[random.randrange(len(string.digits))])
        return ''.join(ret)

    def _send_email(self, address, body, subject, headers=None):
        if headers is None:
            headers = {}
        headers['Subject'] = subject
        headers['From'] = self._email_address
        self.service_messages._send_email(address, body, headers)

    def _calculate_fee(self, result):
        fee = 0

        fees = dict(self._fees)

        # if no conference package, calculate according to participation
        if result['mycode'][-1] == 'y':
            fee += fees[result['mycode']][1]
        else:
            participation = result['mycode'][:4]
            if participation != 'none':
                fee += fees[participation + '-x-n'][1]
                # add the price for the room
                days = DateTime(result['departure']) - DateTime(result['arrival'])
                if result['room'] == 'single':
                    fee += days * fees['board-s'][1]
                elif result['room'] == 'double':
                    fee += days * fees['board-d'][1]
                else:
                    fee += days * fees['voucher'][1]
                    # add external fee too
                    fee += fees['external'][1]
        now = DateTime()
        if now > self._last_pre_date:
            fee += fees['late'][1]

        return fee

    # helper method for the views
    security.declarePublic('set_cache_headers')
    def set_headers(self):
        """sets some HTTP headers to prevent this form from being cached
            and make sure it's presented in the proper codepage"""
        response = self.REQUEST.RESPONSE
        headers = [('Expires', 'Mon, 26 Jul 1997 05:00:00 GMT'),
                    ('Last-Modified', 
                        DateTime("GMT").strftime("%a, %d %b %Y %H:%M:%S GMT")),
                    ('Cache-Control', 'no-cache, must-revalidate'),
                    ('Cache-Control', 'post-check=0, pre-check=0'),
                    ('Pragma', 'no-cache'),
                    ('Content-Type', 'text/html; charset=UTF-8'),
                    ]

        placed = []
        for key, value in headers:
            if key not in placed:
                response.setHeader(key, value)
                placed.append(key)
            else:
                response.addHeader(key, value)

    security.declarePublic('get_show_fields')
    def get_show_fields(self):
        """returns a list of the ids of form fields to show by default"""
        return FormValidator(self.form).show_fields

Globals.InitializeClass(ConferenceService)

manage_addConferenceServiceForm = PageTemplateFile(
    "www/serviceConferenceAdd", globals(),
    __name__='manage_addConferenceServiceForm')

def manage_addConferenceService(self, id, title='', REQUEST=None):
    """Add member message service."""
    object = ConferenceService(id, title)
    self._setObject(id, object)
    object = getattr(self, id)
    add_and_edit(self, id, REQUEST)
    return ''
